/***************************************************************************
 *   Copyright (C) 2008 by Alexey Noskov   *
 *   alexey.noskov@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "kopeteantispamplugin.h"
#include "kopeteantispamconfig.h"

#include <kopete/kopetechatsession.h>
#include <kopete/kopetechatsessionmanager.h>
#include <kopete/kopeteuiglobal.h>
#include <kopete/kopetecontact.h>
#include <kopete/kopetemessage.h>
#include <kopete/kopetemetacontact.h>
#include <kopete/kopetecontactlist.h>
#include <kopete/kopetegroup.h>

#include <kopete/ui/kopeteview.h>

#include <kgenericfactory.h>

#include <qfileinfo.h>
#include <qpixmap.h>

#include <iostream>

static QString notInListGroup = "Not In List";

KopeteAntispamPlugin *KopeteAntispamPlugin::instance = 0;

K_PLUGIN_FACTORY( KopeteAntispamPluginFactory, registerPlugin<KopeteAntispamPlugin>(); )
K_EXPORT_PLUGIN( KopeteAntispamPluginFactory ( "kopete_antispam" ) )

KopeteAntispamPlugin::KopeteAntispamPlugin( QObject *parent, const QVariantList & /*args*/ ) :
	Kopete::Plugin( KopeteAntispamPluginFactory::componentData(), parent ), question( "" )
{
	if ( instance != 0 ) // Нельзя создать второй экземпляр
		return;

	instance = this; // Сохраняем ссылку на экземпляр плагина

	QTimer::singleShot( 1000, this, SLOT( init() ) ); // Инциализируем по таймеру
}

void KopeteAntispamPlugin::init()
{
	connect( Kopete::ChatSessionManager::self() , SIGNAL( aboutToReceive( Kopete::Message & ) ) , this, SLOT( onMessageReceive( Kopete::Message & ) ) );

	connect( this , SIGNAL( settingsChanged() ) , this , SLOT( settingsChanged() ) );

	settingsChanged();
}

void KopeteAntispamPlugin::settingsChanged()
{
	KopeteAntispamConfig * config = KopeteAntispamConfig::self();

	config->readConfig();

	question = config->question();
	passMessage = config->passMessage();
	ignoreUnknown = config->ignoreUnknown();
	removeSpammers = config->removeSpammers();

	updateRegExpList( answers, config->answers() );
    updateRegExpList( whiteList, config->whiteList() );

	#ifdef DEBUG
	std::cerr << "Antispam config changed\n";
	std::cerr << "Antispam question = " << question.data() << "\n";
	std::cerr << "Antispam num answers = " << answers.count() << "\n";
	std::cerr << "Antispam num white listed = " << whiteList.count() << "\n";
	std::cerr << "Antispam answer = " << passMessage.data() << "\n";
	std::cerr << "Antispam ignoreUnknown = " << (ignoreUnknown? "true" : "false") << "\n";
	std::cerr << "Antispam removeSpammers = " << (removeSpammers? "true" : "false") << "\n";
	#endif
}

void KopeteAntispamPlugin::updateRegExpList( QRegExpList & list, const QString & str )
{
	QStringList tokens = str.split( '\n' );

	list.clear();

	for ( QStringList::iterator it = tokens.begin(), end = tokens.end(); it != end; ++ it )
		list.push_back( QRegExp( it->trimmed(), Qt::CaseInsensitive, QRegExp::Wildcard ) ); // Found a token, add regexp it to the vector
}

bool KopeteAntispamPlugin::contactInWhiteList( const Kopete::Contact * contact )
{
	std::cerr << "Checking contact in white list" << std::endl;

	return contact && stringInRegExpList( whiteList, contact->contactId() );
}

bool KopeteAntispamPlugin::stringInRegExpList( const QRegExpList & list, const QString & str )
{
	for ( QRegExpList::const_iterator it = list.begin(); it != list.end(); ++ it ) // For each pattern in regexp list
		if ( it->exactMatch( str ) ) // If pattern matches string
			return true;

	return false;
}

void KopeteAntispamPlugin::removeContact( Kopete::Contact * contact )
{
	Kopete::MetaContact* metaContact = contact->metaContact();

	contact->deleteLater();

	if ( metaContact ) {
		metaContact->removeContact( contact, true );

		if ( metaContact->contacts().count() == 0 ) { // No remaining contacts for metacontact, so remove the MetaContact

			// Get the first group and its members
			Kopete::Group* group = metaContact->groups().first();
			QList<Kopete::MetaContact *> groupMembers = group->members();
			Kopete::ContactList::self()->removeMetaContact( metaContact );

			// Remove the group
			if ( groupMembers.count() == 1 && groupMembers.indexOf( metaContact ) != -1 ) {
				Kopete::ContactList::self()->removeGroup( group );
			}
		}
	}
}

static QList<Kopete::Contact*> contactsToRemove;

void KopeteAntispamPlugin::removeSessionContacts()
{
	for ( QList<Kopete::Contact*>::ConstIterator it = contactsToRemove.begin(); it != contactsToRemove.end(); ++ it )
		removeContact( *it );

	contactsToRemove.clear();
}

bool KopeteAntispamPlugin::contactInList( const Kopete::Contact * contact )
{
	if ( !contact ) // No contact - not in list
		return false;

	const Kopete::MetaContact * mc = contact->metaContact();

	if ( !mc || mc->isTemporary() ) // No metacontact - not in list
		return false;

	for ( QList<Kopete::Group*>::const_iterator it = mc->groups().begin(), e = mc->groups().end(); it != e; ++ it ) // Checking contact groups
		if ( *it && ( (*it)->type() == Kopete::Group::Temporary || (*it)->displayName() == notInListGroup ) ) // Is it in temporary or "Not In List" group?
			return false;

	return true;
}

std::ostream & operator << ( std::ostream & os, const QString & str )
{
	QByteArray ba = str.toUtf8();

	return os << ba.data();
}

//#define DEBUG

void KopeteAntispamPlugin::onMessageReceive( Kopete::Message & msg )
{
	#ifdef DEBUG
	std::cerr << "Message received\n";
	#endif

	if ( !msg.manager() ) {
		std::cerr << "No session!\n";
		return;
	}

	Kopete::ChatSession & session = *msg.manager(); // Get session, in which we receive message

	if ( ( !session.view() ) && // If no view assotiated with session
	     ( !contactInList( msg.from() ) ) && // And contact isn't in contact list
	     ( !contactInWhiteList( msg.from() ) ) ) { // And contact not in white list

		#ifdef DEBUG
		std::cerr << "Checking text. Message is: '" << msg.plainBody() << "'\n";
		#endif

		if ( stringInRegExpList( answers, msg.plainBody() ) && !ignoreUnknown ) { // Message is answer
			#ifdef DEBUG
            std::cerr << "Match!\n";
            #endif

			Kopete::Message msgPass( session.myself(), msg.from() ); // We create message with notification about passing test
			msgPass.setDirection( Kopete::Message::Outbound );
			msgPass.setHtmlBody( passMessage );

			session.sendMessage( msgPass ); // Send message
		} else { // Message isn't answer
			#ifdef DEBUG
            std::cerr << "Mismatch!\n";
            #endif

			if ( !ignoreUnknown ) { // If ignore mode not set, send question to contact
				Kopete::Message msgQuestion( session.myself(), msg.from() ); // We create message with question
				msgQuestion.setDirection( Kopete::Message::Outbound );
				msgQuestion.setHtmlBody( question );

				#ifdef DEBUG
				std::cerr << "Sending question '" << question << "'\n";
				#endif

				if ( removeSpammers ) {
					contactsToRemove.push_back( const_cast<Kopete::Contact*>( msg.from() ) );
					QTimer::singleShot( 10000, this, SLOT( removeSessionContacts() ) );
					//connect( &session, SIGNAL( messageSuccess() ), this, SLOT( removeSessionContacts() ) );
				}

				session.sendMessage( msgQuestion ); // Send message
			} else {
				#ifdef DEBUG
				std::cerr << "Ignoring\n";
				#endif

				if ( removeSpammers ) { // If remove spammers flag is set, remove contact from list
					contactsToRemove.push_back( const_cast<Kopete::Contact*>( msg.from() ) );
					QTimer::singleShot( 10000, this, SLOT( removeSessionContacts() ) );
					//removeContact( (Kopete::Contact*) msg.from() );
					//QTimer::singleShot( 10000, this, SLOT( removeSessionContacts() ) );
				}
			}

			disconnect( &session, SIGNAL( messageAppended( Kopete::Message &, Kopete::ChatSession * ) ), 0, 0 ); // Disconnect session
		}
	}
}
